@@ -93,6 +93,7 @@ See :py:mod:`mrjob.examples` for more examples.
 # since MRJobs need to run in Amazon's generic EMR environment
 from __future__ import with_statement
 
+from copy import copy
 import inspect
 import itertools
 from optparse import Option, OptionParser, OptionGroup, OptionError, OptionValueError
@@ -106,7 +107,7 @@ except ImportError:
 
 # don't use relative imports, to allow this script to be invoked as __main__
 from mrjob.conf import combine_dicts
-from mrjob.parse import parse_mr_job_stderr
+from mrjob.parse import parse_mr_job_stderr, parse_port_range_list, check_kv_pair, check_range_list
 from mrjob.protocol import DEFAULT_PROTOCOL, PROTOCOL_DICT
 from mrjob.runner import CLEANUP_CHOICES, CLEANUP_DEFAULT
 from mrjob.util import log_to_stream, read_input
@@ -118,25 +119,36 @@ def _IDENTITY_MAPPER(key, value):
 # sentinel value; used when running MRJob as a script
 _READ_ARGS_FROM_SYS_ARGV = '_READ_ARGS_FROM_SYS_ARGV'
 
-class SetKeyOption(Option):
 
+class UsageError(Exception):
+    pass
+
+
+class MRJobOptions(Option):
+
+    TYPES = Option.TYPES + ('key_value_pair', 'range_list')
     ACTIONS = Option.ACTIONS + ('set_key',)
     STORE_ACTIONS = Option.STORE_ACTIONS + ('set_key',)
     TYPED_ACTIONS = Option.TYPED_ACTIONS + ('set_key',)
     ALWAYS_TYPED_ACTIONS = Option.ALWAYS_TYPED_ACTIONS + ('set_key',)
 
+    TYPE_CHECKER = copy(Option.TYPE_CHECKER)
+    TYPE_CHECKER["key_value_pair"] = check_kv_pair
+    TYPE_CHECKER["range_list"] = check_range_list
+
     def take_action(self, action, dest, opt, value, values, parser):
         if action == 'set_key':
             try:
-                key, value = value.split('=', 1)
+                store_key, store_value = value
             except ValueError:
                 raise OptionValueError(
-                    "option %s: Expected KEY=VALUE, not %r" % (opt, value))
-            values.ensure_value(dest, {})[key] = value
+                    "option %s: value is not a key_value_pair" % opt)
+            values.ensure_value(dest, {})[store_key] = store_value
         else:
             Option.take_action(
                 self, action, dest, opt, value, values, parser)
 
+
 class MRJob(object):
     """The base class for all MapReduce jobs. See :py:meth:`__init__`
     for details."""
@@ -166,7 +178,7 @@ class MRJob(object):
 
         usage = "usage: %prog [options] [input files]"
         self.option_parser = OptionParser(
-            usage=usage, option_class=SetKeyOption)
+            usage=usage, option_class=MRJobOptions)
         self.configure_options()
 
         # Load and validate options
@@ -351,6 +363,11 @@ class MRJob(object):
 
         :rtype: :py:class:`mrjob.runner.MRJobRunner`
         """
+        bad_words = ('--steps', '--mapper', '--reducer', '--step-num')
+        for w in bad_words:
+            if w in sys.argv:
+                raise UsageError("make_runner() was called with %s. This probably means you tried to use it from __main__, which doesn't work." % w)
+        
         # have to import here so that we can still run the MRJob
         # without importing boto
         from mrjob.emr import EMRJobRunner
@@ -605,138 +622,189 @@ class MRJob(object):
             help='show the steps of mappers and reducers')
 
         # To run mappers or reducers
-        self.option_parser.add_option(
+        self.mux_opt_group = OptionGroup(
+            self.option_parser, 'Running specific parts of the job')
+        self.option_parser.add_option_group(self.mux_opt_group)
+
+        self.mux_opt_group.add_option(
             '--mapper', dest='run_mapper', action='store_true', default=False,
             help='run a mapper')
 
-        self.option_parser.add_option(
+        self.mux_opt_group.add_option(
             '--reducer', dest='run_reducer', action='store_true', default=False,
             help='run a reducer')
 
-        self.option_parser.add_option(
+        self.mux_opt_group.add_option(
             '--step-num', dest='step_num', type='int', default=0,
             help='which step to execute (default is 0)')
 
         # protocol stuff
         protocol_choices = sorted(self.protocols())
+        self.proto_opt_group = OptionGroup(
+            self.option_parser, 'Protocols')
+        self.option_parser.add_option_group(self.proto_opt_group)
 
         self.add_passthrough_option(
-            '-p', '--protocol', dest='protocol',
-            default=self.DEFAULT_PROTOCOL, choices=protocol_choices,
-            help='output protocol for mappers/reducers. Choices: %s (default: %%default)' % ', '.join(protocol_choices))
+            '--input-protocol', dest='input_protocol',
+            opt_group=self.proto_opt_group,
+            default=self.DEFAULT_INPUT_PROTOCOL, choices=protocol_choices,
+            help='protocol to read input with (default: %default)')
+
         self.add_passthrough_option(
             '--output-protocol', dest='output_protocol',
+            opt_group=self.proto_opt_group,
             default=self.DEFAULT_OUTPUT_PROTOCOL,
             choices=protocol_choices,
              help='protocol for final output (default: %s)' % (
             'same as --protocol' if self.DEFAULT_OUTPUT_PROTOCOL is None
             else '%default'))
+
         self.add_passthrough_option(
-            '--input-protocol', dest='input_protocol',
-            default=self.DEFAULT_INPUT_PROTOCOL, choices=protocol_choices,
-            help='protocol to read input with (default: %default)')
+            '-p', '--protocol', dest='protocol',
+            opt_group=self.proto_opt_group,
+            default=self.DEFAULT_PROTOCOL, choices=protocol_choices,
+            help='output protocol for mappers/reducers. Choices: %s (default: %%default)' % ', '.join(protocol_choices))
+
         self.add_passthrough_option(
-			'--strict-protocols', dest='strict_protocols', default=False,
-			action='store_true', help='If something violates an input/output '
-			'protocol then raise an exception')
+            '--strict-protocols', dest='strict_protocols', default=None,
+            opt_group=self.proto_opt_group,
+            action='store_true', help='If something violates an input/output '
+            'protocol then raise an exception')
+
 
         # options for running the entire job
         self.runner_opt_group = OptionGroup(
             self.option_parser, 'Running the entire job')
         self.option_parser.add_option_group(self.runner_opt_group)
 
-        self.runner_opt_group.add_option(
-            '-r', '--runner', dest='runner', default='local',
-            choices=('local', 'hadoop', 'emr', 'inline'),
-            help='Where to run the job: local to run locally, hadoop to run on your Hadoop cluster, emr to run on Amazon ElasticMapReduce. Default is local.')
-        self.runner_opt_group.add_option(
-            '-c', '--conf-path', dest='conf_path', default=None,
-            help='Path to alternate mrjob.conf file to read from')
-        self.runner_opt_group.add_option(
-            '--no-conf', dest='conf_path', action='store_false',
-            help="Don't load mrjob.conf even if it's available")
-        self.runner_opt_group.add_option(
-            '--no-output', dest='no_output',
-            default=False, action='store_true',
-            help="Don't stream output after job completion")
-        self.runner_opt_group.add_option(
-            '--cleanup', dest='cleanup',
-            choices=CLEANUP_CHOICES, default=CLEANUP_DEFAULT,
-            help="when to clean up tmp directories, etc. Choices: %s (default: %%default)" % ', '.join(CLEANUP_CHOICES))
-        self.runner_opt_group.add_option(
-            '-v', '--verbose', dest='verbose', default=False,
-            action='store_true',
-            help='print more messages to stderr')
-        self.runner_opt_group.add_option(
-            '-q', '--quiet', dest='quiet', default=False,
-            action='store_true',
-            help="Don't print anything to stderr")
-        self.runner_opt_group.add_option(
-            '--file', dest='upload_files', action='append',
-            default=[],
-            help='Copy file to the working directory of this script. You can use --file multiple times.')
         self.runner_opt_group.add_option(
             '--archive', dest='upload_archives', action='append',
             default=[],
             help='Unpack archive in the working directory of this script. You can use --archive multiple times.')
 
+        self.runner_opt_group.add_option(
+            '--bootstrap-mrjob', dest='bootstrap_mrjob', action='store_true', default=None,
+            help="Automatically tar up the mrjob library and install it when we run the mrjob. This is the default. Use --no-bootstrap-mrjob if you've already installed mrjob on your Hadoop cluster.")
+
+        self.runner_opt_group.add_option(
+            '-c', '--conf-path', dest='conf_path', default=None,
+            help='Path to alternate mrjob.conf file to read from')
+
+        self.runner_opt_group.add_option(
+            '--cleanup', dest='cleanup',
+            choices=CLEANUP_CHOICES, default=CLEANUP_DEFAULT,
+            help="when to clean up tmp directories, etc. Choices: %s (default: %%default)" % ', '.join(CLEANUP_CHOICES))
+
+        self.runner_opt_group.add_option(
+            '--file', dest='upload_files', action='append',
+            default=[],
+            help='Copy file to the working directory of this script. You can use --file multiple times.')
+
+        self.runner_opt_group.add_option(
+            '--no-bootstrap-mrjob', dest='bootstrap_mrjob', action='store_false', default=None,
+            help="Don't automatically tar up the mrjob library and install it when we run the mrjob. Use this if you've already installed mrjob on your Hadoop cluster.")
+
+        self.runner_opt_group.add_option(
+            '--no-conf', dest='conf_path', action='store_false', default=None,
+            help="Don't load mrjob.conf even if it's available")
+
+        self.runner_opt_group.add_option(
+            '--no-output', dest='no_output',
+            default=None, action='store_true',
+            help="Don't stream output after job completion")
+
         self.runner_opt_group.add_option(
             '-o', '--output-dir', dest='output_dir', default=None,
             help='Where to put final job output. This must be an s3:// URL ' +
-            'for EMR, and an HDFS path for Hadoop, and must be empty')
+            'for EMR, an HDFS path for Hadoop, and a system path for local,' +
+            'and must be empty')
 
         self.runner_opt_group.add_option(
-            '--label', '--job-name-prefix', dest='label', default=None,
-            help='custom prefix for job name, to help us identify the job')
-
-        self.runner_opt_group.add_option(
-            '--owner', dest='owner', default=None,
-            help='custom username to use, to help us identify who ran the job')
-
-        self.runner_opt_group.add_option(
-            '--jobconf', dest='jobconf', default={}, action='set_key',
-            help='-jobconf arg to pass through to hadoop streaming; '
-            'should take the form KEY=VALUE. You can use --jobconf '
-            'multiple times.')
-
-        self.runner_opt_group.add_option(
-            '--cmdenv', dest='cmdenv', default={}, action='set_key',
-            help='set an environment variable for your job inside Hadoop '
-            'streaming. Must take the form KEY=VALUE. You can use --cmdenv '
-            'multiple times.')
-
-        self.runner_opt_group.add_option(
-            '--hadoop-arg', dest='hadoop_extra_args', default=[],
-            action='append', help='Argument of any type to pass to hadoop '
-            'streaming. You can use --hadoop-arg multiple times.')
-
-        self.runner_opt_group.add_option(
-            '--hadoop-input-format', dest='hadoop_input_format', default=None,
-            help='the hadoop InputFormat class used to write the data. '
-            'Custom formats must be included in your hadoop streaming jar (see '
-            '--hadoop-streaming-jar)')
-
-        self.runner_opt_group.add_option(
-            '--hadoop-output-format', dest='hadoop_output_format', default=None,
-            help='the hadoop OutputFormat class used to write the data. '
-            'Custom formats must be included in your hadoop streaming jar (see '
-            '--hadoop-streaming-jar)')
-
-        self.runner_opt_group.add_option(
-            '--hadoop-streaming-jar', dest='hadoop_streaming_jar',
-            default=None,
-            help='Path of your hadoop streaming jar (locally, or on S3/HDFS)')
+            '--python-archive', dest='python_archives', default=[], action='append',
+            help='Archive to unpack and add to the PYTHONPATH of the mr_job script when it runs. You can use --python-archives multiple times.')
 
         self.runner_opt_group.add_option(
             '--python-bin', dest='python_bin', default=None,
             help='Name/path of alternate python binary for mappers/reducers.')
 
+        self.runner_opt_group.add_option(
+            '-q', '--quiet', dest='quiet', default=None,
+            action='store_true',
+            help="Don't print anything to stderr")
+
+        self.runner_opt_group.add_option(
+            '-r', '--runner', dest='runner', default='local',
+            choices=('local', 'hadoop', 'emr', 'inline'),
+            help='Where to run the job: local to run locally, hadoop to run on your Hadoop cluster, emr to run on Amazon ElasticMapReduce. Default is local.')
+
+        self.runner_opt_group.add_option(
+            '--setup-cmd', dest='setup_cmds', action='append',
+            default=[],
+            help='A command to run before each mapper/reducer step in the shell (e.g. "cd my-src-tree; make") specified as a string. You can use --setup-cmd more than once. Use mrjob.conf to specify arguments as a list to be run directly.')
+
+        self.runner_opt_group.add_option(
+            '--setup-script', dest='setup_scripts', action='append',
+            default=[],
+            help='Path to file to be copied into the local working directory and then run. You can use --setup-script more than once. These are run after setup_cmds.')
+
         self.runner_opt_group.add_option(
             '--steps-python-bin', dest='steps_python_bin', default=None,
             help='Name/path of alternate python binary to use to query the '
             'job about its steps, if different from the current Python '
             'interpreter. Rarely needed.')
 
+        self.runner_opt_group.add_option(
+            '-v', '--verbose', dest='verbose', default=None,
+            action='store_true',
+            help='print more messages to stderr')
+
+        # options common to Hadoop and EMR
+        self.hadoop_emr_opt_group = OptionGroup(
+            self.option_parser, 'Running on Hadoop or EMR (these apply when you set -r hadoop or -r emr)')
+        self.option_parser.add_option_group(self.hadoop_emr_opt_group)
+
+        self.hadoop_emr_opt_group.add_option(
+            '--cmdenv', dest='cmdenv', default={}, action='set_key', type='key_value_pair',
+            help='set an environment variable for your job inside Hadoop '
+            'streaming. Must take the form KEY=VALUE. You can use --cmdenv '
+            'multiple times.')
+
+        self.hadoop_emr_opt_group.add_option(
+            '--hadoop-arg', dest='hadoop_extra_args', default=[],
+            action='append', help='Argument of any type to pass to hadoop '
+            'streaming. You can use --hadoop-arg multiple times.')
+
+        self.hadoop_emr_opt_group.add_option(
+            '--hadoop-input-format', dest='hadoop_input_format', default=None,
+            help='the hadoop InputFormat class used to write the data. '
+            'Custom formats must be included in your hadoop streaming jar (see '
+            '--hadoop-streaming-jar)')
+
+        self.hadoop_emr_opt_group.add_option(
+            '--hadoop-output-format', dest='hadoop_output_format', default=None,
+            help='the hadoop OutputFormat class used to write the data. '
+            'Custom formats must be included in your hadoop streaming jar (see '
+            '--hadoop-streaming-jar)')
+
+        self.hadoop_emr_opt_group.add_option(
+            '--hadoop-streaming-jar', dest='hadoop_streaming_jar',
+            default=None,
+            help='Path of your hadoop streaming jar (locally, or on S3/HDFS)')
+
+        self.hadoop_emr_opt_group.add_option(
+            '--jobconf', dest='jobconf', default={}, action='set_key', type='key_value_pair',
+            help='-jobconf arg to pass through to hadoop streaming; '
+            'should take the form KEY=VALUE. You can use --jobconf '
+            'multiple times.')
+
+        self.hadoop_emr_opt_group.add_option(
+            '--label', '--job-name-prefix', dest='label', default=None,
+            help='custom prefix for job name, to help us identify the job')
+
+        self.hadoop_emr_opt_group.add_option(
+            '--owner', dest='owner', default=None,
+            help='custom username to use, to help us identify who ran the job')
+
         # options for running the job on Hadoop
         self.hadoop_opt_group = OptionGroup(
             self.option_parser, 'Running on Hadoop (these apply when you set -r hadoop)')
@@ -745,6 +813,7 @@ class MRJob(object):
         self.hadoop_opt_group.add_option(
             '--hadoop-bin', dest='hadoop_bin', default=None,
             help='hadoop binary. Defaults to $HADOOP_HOME/bin/hadoop')
+
         self.hadoop_opt_group.add_option(
             '--hdfs-scratch-dir', dest='hdfs_scratch_dir',
             default=None,
@@ -756,47 +825,134 @@ class MRJob(object):
         self.option_parser.add_option_group(self.emr_opt_group)
 
         self.emr_opt_group.add_option(
-            '--ec2-instance-type', dest='ec2_instance_type', default=None,
-            help='Type of EC2 instance(s) to launch (e.g. m1.small, c1.xlarge, m2.xlarge). See http://aws.amazon.com/ec2/instance-types/ for the full list.')
+            '--aws-availability-zone', dest='aws_availability_zone', default=None,
+            help='Availability zone to run the job flow on')
+
         self.emr_opt_group.add_option(
-            '--ec2-master-instance-type', dest='ec2_master_instance_type', default=None,
-            help='Type of EC2 instance for master node only')
+            '--aws-region', dest='aws_region', default=None,
+            help='Region to connect to S3 and EMR on (e.g. us-west-1).')
+
         self.emr_opt_group.add_option(
-            '--ec2-slave-instance-type', dest='ec2_slave_instance_type', default=None,
-            help='Type of EC2 instance for slave nodes only')
+            '--bootstrap-action', dest='bootstrap_actions', action='append',
+            default=[],
+            help='Raw bootstrap action scripts to run before any of the other bootstrap steps. You can use --bootstrap-action more than once. Local scripts will be automatically uploaded to S3. To specify arguments, just use quotes: "foo.sh arg1 arg2"')
+
         self.emr_opt_group.add_option(
-            '--num-ec2-instances', dest='num_ec2_instances', default=None,
-            type='int',
-            help='Number of EC2 instances to launch')
+            '--bootstrap-cmd', dest='bootstrap_cmds', action='append',
+            default=[],
+            help='Commands to run on the master node to set up libraries, etc. You can use --bootstrap-cmd more than once. Use mrjob.conf to specify arguments as a list to be run directly.')
+
         self.emr_opt_group.add_option(
-            '--s3-scratch-uri', dest='s3_scratch_uri', default=None,
-            help='URI on S3 to use as our temp directory.')
+            '--bootstrap-file', dest='bootstrap_files', action='append',
+            default=[],
+            help='File to upload to the master node before running bootstrap_cmds (for example, debian packages). These will be made public on S3 due to a limitation of the bootstrap feature. You can use --bootstrap-file more than once.')
+
         self.emr_opt_group.add_option(
-            '--s3-log-uri', dest='s3_log_uri', default=None,
-            help='URI on S3 to write logs into')
+            '--bootstrap-python-package', dest='bootstrap_python_packages', action='append',
+            default=[],
+            help='Path to a Python module to install on EMR. These should be standard python module tarballs where you can cd into a subdirectory and run ``sudo python setup.py install``. You can use --bootstrap-python-package more than once.')
+
+        self.emr_opt_group.add_option(
+            '--bootstrap-script', dest='bootstrap_scripts', action='append',
+            default=[],
+            help='Script to upload and then run on the master node (a combination of bootstrap_cmds and bootstrap_files). These are run after the command from bootstrap_cmds. You can use --bootstrap-script more than once.')
+
         self.emr_opt_group.add_option(
             '--check-emr-status-every', dest='check_emr_status_every',
             default=None, type='int',
             help='How often (in seconds) to check status of your EMR job')
+
         self.emr_opt_group.add_option(
-            '--ssh-tunnel-to-job-tracker', dest='ssh_tunnel_to_job_tracker',
-            default=None, action='store_true',
-            help='Open up an SSH tunnel to the Hadoop job tracker')
+            '--ec2-instance-type', dest='ec2_instance_type', default=None,
+            help='Type of EC2 instance(s) to launch (e.g. m1.small, c1.xlarge, m2.xlarge). See http://aws.amazon.com/ec2/instance-types/ for the full list.')
+
         self.emr_opt_group.add_option(
-            '--ssh-tunnel-is-open', dest='ssh_tunnel_is_open',
-            default=None, action='store_true',
-            help='Make ssh tunnel accessible from remote hosts (not just localhost)')
+            '--ec2-key-pair', dest='ec2_key_pair', default=None,
+            help='Name of the SSH key pair you set up for EMR')
+
+        self.emr_opt_group.add_option(
+            '--ec2-key-pair-file', dest='ec2_key_pair_file', default=None,
+            help='Path to file containing SSH key for EMR')
+
+        self.emr_opt_group.add_option(
+            '--ec2-master-instance-type', dest='ec2_master_instance_type', default=None,
+            help='Type of EC2 instance for master node only')
+
+        self.emr_opt_group.add_option(
+            '--ec2-slave-instance-type', dest='ec2_slave_instance_type', default=None,
+            help='Type of EC2 instance for slave nodes only')
+
+        self.emr_opt_group.add_option(
+            '--emr-endpoint', dest='emr_endpoint', default=None,
+            help='Optional host to connect to when communicating with S3 (e.g. us-west-1.elasticmapreduce.amazonaws.com). Default is to infer this from aws_region.')
+
         self.emr_opt_group.add_option(
             '--emr-job-flow-id', dest='emr_job_flow_id', default=None,
             help='ID of an existing EMR job flow to use')
+
+        self.emr_opt_group.add_option(
+            '--enable-emr-debugging', dest='enable_emr_debugging', default=None,
+            action='store_true',
+            help='Enable storage of Hadoop logs in SimpleDB')
+
+        self.emr_opt_group.add_option(
+            '--disable-emr-debugging', dest='enable_emr_debugging', action='store_false',
+            help='Enable storage of Hadoop logs in SimpleDB')
+
         self.emr_opt_group.add_option(
             '--hadoop-streaming-jar-on-emr', dest='hadoop_streaming_jar_on_emr',
             default=None,
             help='Local path of the hadoop streaming jar on the EMR node. Rarely necessary')
+
         self.emr_opt_group.add_option(
             '--hadoop-version', dest='hadoop_version', default=None,
             help='Version of Hadoop to spin up on EMR. Default is 0.18, but will change to 0.20 in v0.3.0 of mrjob.')
 
+        self.emr_opt_group.add_option(
+            '--num-ec2-instances', dest='num_ec2_instances', default=None,
+            type='int',
+            help='Number of EC2 instances to launch')
+
+        self.emr_opt_group.add_option(
+            '--s3-endpoint', dest='s3_endpoint', default=None,
+            help='Host to connect to when communicating with S3 (e.g. s3-us-west-1.amazonaws.com). Default is to infer this from aws_region.')
+
+        self.emr_opt_group.add_option(
+            '--s3-log-uri', dest='s3_log_uri', default=None,
+            help='URI on S3 to write logs into')
+
+        self.emr_opt_group.add_option(
+            '--s3-scratch-uri', dest='s3_scratch_uri', default=None,
+            help='URI on S3 to use as our temp directory.')
+
+        self.emr_opt_group.add_option(
+            '--s3-sync-wait-time', dest='s3_sync_wait_time', default=None, type='float',
+            help='How long to wait for S3 to reach eventual consistency. This is typically less than a second (zero in us-west) but the default is 5.0 to be safe.')
+
+        self.emr_opt_group.add_option(
+            '--ssh-bin', dest='ssh_bin', default=None,
+            help='Path to the ssh binary. Defaults to ssh.')
+
+        self.emr_opt_group.add_option(
+            '--ssh-bind-ports', dest='ssh_bind_ports', default=None,
+            type='range_list',
+            help='A list of port ranges that are safe to listen on, delimited by colons and commas with the syntax 2000[:2001][,2003,2005:2008,etc]. Defaults to 40001:40840.')
+
+        self.emr_opt_group.add_option(
+            '--ssh-tunnel-is-closed', dest='ssh_tunnel_is_open',
+            default=None, action='store_false',
+            help='Make ssh tunnel accessible from localhost only')
+
+        self.emr_opt_group.add_option(
+            '--ssh-tunnel-is-open', dest='ssh_tunnel_is_open',
+            default=None, action='store_true',
+            help='Make ssh tunnel accessible from remote hosts (not just localhost)')
+
+        self.emr_opt_group.add_option(
+            '--ssh-tunnel-to-job-tracker', dest='ssh_tunnel_to_job_tracker',
+            default=None, action='store_true',
+            help='Open up an SSH tunnel to the Hadoop job tracker')
+
 
     def add_passthrough_option(self, *args, **kwargs):
         """Function to create options which both the job runner
@@ -812,7 +968,10 @@ class MRJob(object):
         If you want to pass files through to the mapper/reducer, use
         :py:meth:`add_file_option` instead.
         """
-        pass_opt = self.option_parser.add_option(*args, **kwargs)
+        if not kwargs.has_key('opt_group'):
+            pass_opt = self.option_parser.add_option(*args, **kwargs)
+        else:
+            pass_opt = kwargs.pop('opt_group').add_option(*args, **kwargs)
 
         # We only support a subset of option parser actions
         SUPPORTED_ACTIONS = (
@@ -910,6 +1069,7 @@ class MRJob(object):
         want to add or change lots of keyword arguments.
         """
         return {
+            'bootstrap_mrjob': self.options.bootstrap_mrjob,
             'cleanup': self.options.cleanup,
             'cmdenv': self.options.cmdenv,
             'conf_path': self.options.conf_path,
@@ -925,7 +1085,10 @@ class MRJob(object):
             'label': self.options.label,
             'output_dir': self.options.output_dir,
             'owner': self.options.owner,
+            'python_archives': self.options.python_archives,
             'python_bin': self.options.python_bin,
+            'setup_cmds': self.options.setup_cmds,
+            'setup_scripts': self.options.setup_scripts,
             'stdin': self.stdin,
             'steps_python_bin': self.options.steps_python_bin,
             'upload_archives': self.options.upload_archives,
@@ -993,7 +1156,7 @@ class MRJob(object):
         hadoop or executed via subprocess.
 
         These are passed to :py:meth:`mrjob.runner.MRJobRunner.__init__`
-        as ``file_upload_args``.
+        as ``extra_args``.
         """
         master_option_dict = self.options.__dict__
 
@@ -1026,7 +1189,7 @@ class MRJob(object):
         of tuples of ``('--argname', path)``
 
         These are passed to :py:meth:`mrjob.runner.MRJobRunner.__init__`
-        as ``extra_args``.
+        as ``file_upload_args``.
         """
         file_upload_args = []
 
@@ -1059,7 +1222,7 @@ class MRJob(object):
 
             @classmethod
             def protocols(cls):
-                protocol_dict = super(MRYourJob, self).protocols()
+                protocol_dict = super(MRYourJob, cls).protocols()
                 protocol_dict['rot13'] = Rot13Protocol
                 return protocol_dict
 
